import json
import logging
import os
import re
import unittest

from pararamio import Pararamio


class SensitiveFormatter(logging.Formatter):
    """Formatter that removes sensitive information in urls."""

    @staticmethod
    def _filter(s):
        return re.sub(r'password":\s*\"\w+\"', r'password": "*******"', s)

    def format(self, record):
        original = logging.Formatter.format(self, record)
        return self._filter(original)


DOTENV_PATH = os.path.join(os.path.dirname(__file__), '../..', '.env')


class BasePararamioTest(unittest.TestCase):
    def setUp(self):
        if os.path.exists(DOTENV_PATH):
            with open(DOTENV_PATH, 'r') as f:
                for line in f.readlines():
                    key, value = line.strip().split('=', 1)
                    os.environ[key] = value
        self.user1 = json.loads(os.environ.get('PARARAMIO_FIRST_USER', '{}'))
        self.user2 = json.loads(os.environ.get('PARARAMIO_SECOND_USER', '{}'))
        if not self.user1 or not self.user2:
            raise self.skipTest('Please provide PARARAMIO_FIRST_USER and PARARAMIO_SECOND_USER env variables')
        self.client = Pararamio(**self.user1, cookie_path='test.cookie')
        self.client2 = Pararamio(**self.user2, cookie_path='test.cookie2')
        self.__post = None
        if bool(os.environ.get('DEBUG', False)):
            logger = logging.getLogger('pararamio')
            logger.setLevel(logging.DEBUG)
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            ch.setFormatter(SensitiveFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
            logger.addHandler(ch)



from typing import List, Dict, Callable, Any
from datetime import datetime
from enum import Enum

from pararamio.utils import parse_iso_datetime

__all__ = ('ActivityAction', 'Activity')


class ActivityAction(Enum):
    ONLINE = 'online'
    OFFLINE = 'offline'
    AWAY = 'away'
    READ = 'thread-read'
    POST = 'thread-post'
    CALL = 'calling'
    CALL_END = 'endcall'


class Activity:
    action: ActivityAction
    time: datetime

    def __init__(self, action: ActivityAction, time: datetime):
        self.action = action
        self.time = time

    def __str__(self):
        return str((self.time, self.action))

    @classmethod
    def _from_api_data(cls, data: Dict[str, str]) -> 'Activity':
        return cls(action=ActivityAction(data['action']), time=parse_iso_datetime(data, 'datetime'))

    @classmethod
    def get_activity(cls, page_loader: Callable[..., Dict[str, Any]], start: datetime, end: datetime, actions: List[ActivityAction] = None) -> List['Activity']:
        results = []
        if not actions:
            actions = [None]
        for action in actions:
            page = 1
            is_last_page = False
            while not is_last_page:
                data = page_loader(action, page=page).get('data', [])
                if not data:
                    break
                for d in data:
                    act = Activity._from_api_data(d)
                    if act.time > end:
                        continue
                    if act.time < start:
                        is_last_page = True
                        break
                    results.append(act)
                page += 1
        return sorted(results, key=lambda x: x.time)



from typing import TYPE_CHECKING, Any, Callable, Dict, List

from pararamio.exceptions import (
    PararamioRequestException,
    PararamioServerResponseException,
    PararamioValidationException,
)

if TYPE_CHECKING:
    from pararamio.client import Pararamio
    from .chat import Chat

__all__ = ('Poll',)


ATTR_FORMATTERS: Dict[str, Callable[[dict, str], Any]] = {}


class PollOption:
    """
    Represents an option within a poll.

    :ivar id: The unique identifier for the poll option.
    :vartype id: int
    :ivar text: The text description of the poll option.
    :vartype text: str
    :ivar count: The number of votes received for this poll option.
    :vartype count: int
    :ivar vote_users: A list of user identifiers who voted for this option.
    :vartype vote_users: List[int]
    """
    id: int
    text: str
    count: int
    vote_users: List[int]

    def __init__(self, id: int, text: str, count: int, vote_users: List[int]) -> None:
        self.id = id
        self.text = text
        self.count = count
        self.vote_users = vote_users

    @classmethod
    def from_response_data(cls, data: Dict[str, Any]) -> 'PollOption':
        """
        Create a PollOption object from the response data.

        :param data: A dictionary containing the response data.
        :type data: Dict[str, Any]
        :return: An instance of PollOption class.
        :rtype: PollOption
        :raises PararamioServerResponseException: If any required field is missing in the response data.
        """
        for field in cls.__annotations__:
            if field not in data:
                raise PararamioServerResponseException(
                    f'invalid server vote option response, missing {field}',
                    data,
                )
        return cls(**data)


class Poll:
    """
    Represents a poll object.
    :ivar vote_uid: A unique identifier for the poll.
    :vartype vote_uid: str
    :ivar chat_id: The identifier of the chat where the poll is conducted.
    :vartype chat_id: int
    :ivar anonymous: A boolean indicating whether the poll is anonymous or not.
    :vartype anonymous: bool
    :ivar mode: Options select mode of the poll ('one' for single or 'more' for multi).
    :vartype mode: str
    :ivar options: A list of PollOption objects representing the available poll options.
    :vartype options: List[PollOption]
    :ivar question: The question of the poll.
    :vartype question: str
    :ivar total_user: The total number of users who participated in the poll.
    :vartype total_user: int
    :ivar total_answer: The total number of responses received in the poll.
    :vartype total_answer: int
    :ivar user_id: The identifier of the user who created the poll.
    :vartype user_id: int
    """
    _client: 'Pararamio'
    _data: Dict[str, Any]

    vote_uid: str
    chat_id: int
    anonymous: bool
    mode: str
    options: List[PollOption]
    question: str
    total_user: int
    total_answer: int
    user_id: int

    def __init__(self, client: 'Pararamio', vote_uid: str, **kwargs) -> None:
        self._client = client
        self.vote_uid = vote_uid
        self._data = {**kwargs, 'vote_uid': vote_uid}

    def __getattr__(self, name: str) -> Any:
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        return self._data[name]

    def __str__(self) -> str:
        return self.question

    @property
    def client(self) -> 'Pararamio':
        return self._client

    def _update(self, response: Dict[str, Any]) -> 'Poll':
        """
        Update the Poll object with the response data.

        :param response: A dictionary containing the response data.
        :type response: Dict[str, Any]
        :return: The updated Poll object.
        :rtype: Poll
        :raises PararamioServerResponseException: If 'vote' key is not present in the response.
        """
        if 'vote' not in response:
            raise PararamioServerResponseException(
                f'failed to load data for vote {self.vote_uid} in chat {self._chat.id}',
                response,
            )
        self._data = {
            k: v if k != 'options' else [PollOption.from_response_data(opt) for opt in v]
            for k, v in response['vote'].items()
        }
        return self

    def load(self) -> 'Poll':
        """
        Load the poll's data from the pararam server.

        :return: The updated instance of the poll.
        :rtype: Poll
        """
        res = self._client.api_get(f'/msg/vote/{self.vote_uid}')
        return self._update(res)

    @classmethod
    def create(cls, chat: 'Chat', question: str, mode: str, anonymous: bool, options: List[str]) -> 'Poll':
        """
        Create a new poll in the specified pararam chat.

        :param chat: The chat in which the poll will be created.
        :type chat: Chat
        :param question: The question for the poll.
        :type question: str
        :param mode: Options select mode of the poll ('one' for single or 'more' for multi).
        :type mode: str
        :param anonymous: Whether the poll should be anonymous or not.
        :type anonymous: bool
        :param options: The list of options for the poll.
        :type options: List[str]
        :return: The created Poll object.
        :rtype: Poll
        :raises PararamioRequestException: If the request to create the poll fails.
        """
        res = chat.client.api_post(
            '/msg/vote',
            {
                'chat_id': chat.id,
                'question': question,
                'options': options,
                'mode': mode,
                'anonymous': anonymous,
            }
        )
        if not res:
            raise PararamioRequestException('Failed to create post')
        return cls(chat.client, res['vote_uid']).load()

    def _vote(self, option_ids: List[int]) -> 'Poll':
        """
        Vote on the poll by selecting the given option IDs.

        :param option_ids: A list of integers representing the IDs of the options to vote for.
        :type option_ids: List[int]
        :return: The updated Poll object after voting.
        :rtype: Poll
        :raises PararamioValidationException: If any of the option IDs are incorrect.
        """
        ids_ = [opt.id for opt in self.options]
        if not all(opt_id in ids_ for opt_id in option_ids):
            raise PararamioValidationException('incorrect option')
        res = self._client.api_put(
            f'/msg/vote/{self.vote_uid}',
            {
                'variants': option_ids,
            }
        )
        return self._update(res)

    def vote(self, option_id: int) -> 'Poll':
        """
        Vote for a specific option in the poll.

        :param option_id: The ID of the option to vote for.
        :type option_id: int
        :return: The updated Poll object after voting.
        :rtype: Poll
        :raises PararamioValidationException: If the option_id is invalid.
        """
        return self._vote([option_id])

    def vote_multi(self, option_ids: List[int]) -> 'Poll':
        """
        Vote for multiple options in a poll.

        :param option_ids: A list of integers representing the IDs of the options to vote for.
        :type option_ids: List[int]
        :return: The updated instance of the poll.
        :rtype: Poll
        :raises PararamioValidationException: If the poll mode is not 'more' or if any of the option IDs are incorrect.
        """
        if self.mode != 'more':
            raise PararamioValidationException(f'incorrect poll mode ({self.mode}) for multi voting')
        return self._vote(option_ids)

    def retract(self) -> 'Poll':
        """
        Retracts the vote from the poll.

        :return: The updated instance of the poll.
        :rtype: Poll
        """
        return self._vote([])



# noqa
import unittest

from pararamio.utils.helpers import lazy_loader


# noinspection PyTypeChecker
class UtilsTestCase(unittest.TestCase):
    def test_001_lazy_loader(self):
        items = list(range(0, 50))
        calls_count = 0

        def load_more(_, items):
            nonlocal calls_count
            calls_count += 1
            return items

        self.assertListEqual(items, list(lazy_loader({}, items, load_more, per_load=25)), 'returned lists not equals')
        self.assertEqual(calls_count, 2, 'Calls count is not equal 2')

if __name__ == '__main__':
    unittest.main()



import base64
import binascii
import hashlib
import hmac
import json
import logging
import time
from datetime import datetime
from http.cookiejar import CookieJar
from typing import Any, Dict, Tuple
from urllib.error import HTTPError

from pararamio._types import HeaderLikeT, SecondStepFnT
from pararamio.constants import XSRF_HEADER_NAME
import pararamio.exceptions as ex
from pararamio.utils.requests import api_request, raw_api_request


__all__ = ('get_xsrf_token', 'authenticate', 'do_second_step', 'do_second_step_with_code',)

XSFR_URL = INIT_URL = '/auth/init'
LOGIN_URL = '/auth/login/password'
TWO_STEP_URL = '/auth/totp'
AUTH_URL = '/auth/next'
log = logging.getLogger('pararamio')


def get_xsrf_token(cookie_jar: CookieJar) -> str:
    _, headers = raw_api_request(XSFR_URL, cookie_jar=cookie_jar)
    for key, value in headers:
        if key.lower() == 'x-xsrftoken':
            return value
    raise ex.PararamioXSFRRequestError('XSFR Header was not found in %s url' % XSFR_URL)


def do_init(cookie_jar: CookieJar, headers: dict) -> Tuple[bool, dict]:
    try:
        return True, api_request(
            INIT_URL,
            method='GET',
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise


def do_login(login: str, password: str, cookie_jar: CookieJar, headers: dict) -> Tuple[bool, dict]:
    try:
        return True, api_request(
            LOGIN_URL,
            method='POST',
            data={'email': login, 'password': password},
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise


def do_taking_secret(cookie_jar: CookieJar, headers: dict) -> Tuple[bool, dict]:
    try:
        return True, api_request(
            AUTH_URL,
            method='GET',
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise


def do_second_step(cookie_jar: CookieJar, headers: dict, key: str) -> Tuple[bool, Dict[str, str]]:
    """
    do second step pararam login with TFA key or raise Exception
    :param cookie_jar: cookie container
    :param headers: headers to send
    :param key: key to generate one time code
    :return: True if login success
    """
    if not key:
        raise ex.PararamioSecondFactorAuthenticationException('key can not be empty')
    try:
        key = generate_otp(key)
    except binascii.Error:
        raise ex.PararamioSecondFactorAuthenticationException('Invalid second step key')
    try:
        resp = api_request(
            TWO_STEP_URL,
            method='POST',
            data={'code': key},
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise
    return True, resp


def do_second_step_with_code(cookie_jar: CookieJar, headers: Dict[str, str], code: str) -> Tuple[bool, Dict[str, str]]:
    """
    do second step pararam login with TFA code or raise Exception
    :param cookie_jar: cookie container
    :param headers: headers to send
    :param code: 6 digits code
    :return:  True if login success
    """
    if not code:
        raise ex.PararamioSecondFactorAuthenticationException('code can not be empty')
    if len(code) != 6:
        raise ex.PararamioSecondFactorAuthenticationException('code must be 6 digits len')
    try:
        resp = api_request(
            TWO_STEP_URL,
            method='POST',
            data={'code': code},
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise
    return True, resp


def authenticate(login: str,
                 password: str,
                 cookie_jar: CookieJar,
                 headers: HeaderLikeT = None,
                 second_step_fn: SecondStepFnT = do_second_step,
                 second_step_arg: str = None
                 ) -> Tuple[bool, Dict[str, Any], str]:
    if not headers or XSRF_HEADER_NAME not in headers:
        if headers is None:
            headers = {}
        headers[XSRF_HEADER_NAME] = get_xsrf_token(cookie_jar)

    success, resp = do_login(login, password, cookie_jar, headers)

    if resp.get('codes', {}).get('non_field', '') == 'captcha_required':
        try:
            from pararamio.utils.captcha import show_captcha

            success = show_captcha(f'login:{login}', headers, cookie_jar)
            if not success:
                raise ex.PararamioCaptchaAuthenticationException('Captcha required')
            else:
                success, resp = do_login(login, password, cookie_jar, headers)
        except ImportError:
            raise ex.PararamioCaptchaAuthenticationException('Captcha required, but exception when show it')

    if not success and resp.get('error', 'xsrf'):
        log.debug('invalid xsrf trying to get new one')
        headers[XSRF_HEADER_NAME] = get_xsrf_token(cookie_jar)
        success, resp = do_login(login, password, cookie_jar, headers)

    if not success:
        log.error('authentication failed: %s', resp.get('error', ''))
        raise ex.PararamioPasswordAuthenticationException('Login, password authentication failed')

    if second_step_fn and second_step_arg:
        success, resp = second_step_fn(cookie_jar, headers, second_step_arg)
        if not success:
            raise ex.PararamioSecondFactorAuthenticationException('Second factor authentication failed')

    success, resp = do_taking_secret(cookie_jar, headers)
    if not success:
        raise ex.PararamioAuthenticationException('Taking secret failed')

    success, resp = do_init(cookie_jar, headers)
    return True, {'user_id': resp.get('user_id')}, headers[XSRF_HEADER_NAME]


def generate_otp(key: str) -> str:
    digits = 6
    digest = hashlib.sha1
    interval = 30

    def byte_secret(s):
        missing_padding = len(s) % 8
        if missing_padding != 0:
            s += '=' * (8 - missing_padding)
        return base64.b32decode(s, casefold=True)

    def int_to_byte_string(i, padding=8):
        result = bytearray()
        while i != 0:
            result.append(i & 0xFF)
            i >>= 8
        return bytes(bytearray(reversed(result)).rjust(padding, b'\0'))

    def time_code(for_time):
        i = time.mktime(for_time.timetuple())
        return int(i / interval)

    hmac_hash = hmac.new(
        byte_secret(key),
        int_to_byte_string(time_code(datetime.now())),
        digest,
    ).digest()

    hmac_hash = bytearray(hmac_hash)
    offset = hmac_hash[-1] & 0xf
    code = ((hmac_hash[offset] & 0x7f) << 24 | (hmac_hash[offset + 1] & 0xff) << 16 | (hmac_hash[offset + 2] & 0xff) << 8 | (hmac_hash[offset + 3] & 0xff))
    str_code = str(code % 10 ** digits)
    while len(str_code) < digits:
        str_code = '0' + str_code

    return str_code



import sys
from http.cookiejar import CookieJar
from typing import Any, Callable, Dict, Tuple, TypeVar, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict, Optional, TypeVar

    if sys.version_info >= (3, 11):
        from typing import NotRequired
        class ProfileTypeT(TypedDict):
            unique_name: str
            id: int
            info: Optional[str]
            find_strict: bool
            name: str
            is_google: bool
            two_step_enabled: bool
            has_password: bool
            phoneconfirmed: bool
            email: str
            phonenumber: NotRequired[Optional[str]]


        class TextParsedT(TypedDict):
            type: str
            value: str
            name: NotRequired[str]
            id: NotRequired[int]


        class PostMetaUserT(TypedDict):
            id: int
            name: str
            unique_name: str
            is_bot: NotRequired[bool]

    else:
        class ProfileTypeT(TypedDict):
            unique_name: str
            id: int
            info: Optional[str]
            find_strict: bool
            name: str
            is_google: bool
            two_step_enabled: bool
            has_password: bool
            phoneconfirmed: bool
            email: str
            phonenumber: Optional[str]


        class TextParsedT(TypedDict):
            type: str
            value: str
            name: Optional[str]
            id: Optional[int]


        class PostMetaUserT(TypedDict):
            id: int
            name: str
            unique_name: str
            is_bot: Optional[bool]


    class PostMetaFileT(TypedDict):
        name: str
        guid: str
        size: int
        mime_type: str
        origin: Tuple[int, int]


    class MetaReplyT(TypedDict):
        text: str
        user_id: int
        user_name: str
        in_thread_no: int


    class PostMetaThreadT(TypedDict):
        title: str


    class PostMetaT(TypedDict):
        user: PostMetaUserT
        thread: PostMetaThreadT
        file: PostMetaFileT
        reply: MetaReplyT


    class PostMention(TypedDict):
        id: int
        name: str
        value: str

else:
    ProfileTypeT = Dict[str, Any]
    PostMetaT = Dict[str, Any]
    PostMention = Dict[str, Union[str, int]]
    TextParsedT = Dict[str, Union[str, int]]

CookieJarT = TypeVar('CookieJarT', bound=CookieJar)
QuoteRangeT = Dict[str, Union[str, int]]
HeaderLikeT = Dict[str, str]
SecondStepFnT = Callable[[CookieJar, Dict[str, str], str], Tuple[bool, Dict[str, str]]]



import logging
import mimetypes
import os
from http.cookiejar import CookieJar, FileCookieJar, LoadError, MozillaCookieJar
from io import BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Callable, Dict, Iterable, List, Optional, Tuple, Union

from pararamio._types import ProfileTypeT, SecondStepFnT
from pararamio.chat import Chat
from pararamio.constants import XSRF_HEADER_NAME
from pararamio.exceptions import PararamioAuthenticationException, PararamioHTTPRequestException, PararamioValidationException
from pararamio.file import File
from pararamio.group import Group
from pararamio.post import Post
from pararamio.user import User
from pararamio.utils.authentication import authenticate, do_second_step, do_second_step_with_code, get_xsrf_token
from pararamio.utils.helpers import check_login_opts, get_empty_vars, lazy_loader, unescape_dict
from pararamio.utils.requests import api_request, delete_file, download_file, xupload_file

__all__ = ('Pararamio',)
log = logging.getLogger('pararamio.client')


class Pararamio:
    _login: Optional[str]
    _password: Optional[str]
    _key: Optional[str]
    _authenticated: bool
    _cookie: Union[CookieJar, FileCookieJar]
    __profile: Optional[dict]
    __headers: Dict[str, str]
    __user: dict

    def __init__(self, login: str = None, password: str = None, key: str = None, cookie: CookieJar = None, cookie_path: str = None,
                 ignore_broken_cookie: bool = False):
        self._login = login
        self._password = password
        self._key = key
        self.__headers = {}
        self.__profile = None
        self.__user = {}
        self._authenticated = False
        if cookie is not None:
            self._cookie = cookie
        elif cookie_path is not None:
            self._cookie = MozillaCookieJar(cookie_path)
            if os.path.exists(cookie_path):
                if not os.path.isfile(cookie_path):
                    raise OSError(f'path {cookie_path} is directory')
                if not os.access(cookie_path, os.R_OK):
                    raise OSError(f'file {cookie_path} is not readable')
                if not os.access(cookie_path, os.W_OK):
                    raise OSError(f'file {cookie_path} is not writable')
                try:
                    self._cookie.load(ignore_discard=True)
                    self._authenticated = True
                except LoadError as e:
                    log.error('failed to load cookie from file %s', cookie_path)
                    if not ignore_broken_cookie:
                        raise OSError(e)
        else:
            self._cookie = CookieJar()
        for cj in self._cookie:
            if cj.name == '_xsrf':
                self.__headers[XSRF_HEADER_NAME] = str(cj.value)
                break

    @property
    def cookies(self) -> Union[CookieJar, FileCookieJar]:
        if not self._authenticated:
            self.authenticate()
        return self._cookie

    @property
    def headers(self) -> Dict[str, str]:
        if not self._authenticated:
            self.authenticate()
        return self.__headers

    def _save_cookie(self) -> None:
        if isinstance(self._cookie, FileCookieJar):
            self._cookie.save(ignore_discard=True)

    def _profile(self, raise_on_error: bool = False) -> 'ProfileTypeT':
        return unescape_dict(self.api_get('/user/me', raise_on_error=raise_on_error), keys=['name'])

    def _do_auth(
        self, login: str, password: str, cookie_jar: CookieJar, headers: Dict[str, str], second_step_fn: SecondStepFnT,
        second_step_arg: str, ) -> None:
        self._authenticated, user, xsrf = authenticate(login, password, cookie_jar, headers, second_step_fn, second_step_arg)
        if self._authenticated:
            self.__user = user
            self.__headers[XSRF_HEADER_NAME] = xsrf
            self._save_cookie()

    def _authenticate(self, second_step_fn: SecondStepFnT, second_step_arg: str, login: str = None, password: str = None) -> bool:
        login = login or self._login
        password = password or self._password
        if not check_login_opts(login, password):
            raise PararamioAuthenticationException(f'{get_empty_vars(login=login, password=password)} must be set and not empty')
        if not self._cookie:
            self._do_auth(login, password, self._cookie, self.__headers, second_step_fn, second_step_arg)
        try:
            self._authenticated = True
            self._profile(raise_on_error=True)
        except PararamioHTTPRequestException:
            self._authenticated = False
            self._do_auth(login, password, self._cookie, self.__headers, second_step_fn, second_step_arg)
        return self._authenticated

    def authenticate(self, login: str = None, password: str = None, key: str = None) -> bool:
        """
        Authenticate client with totp key
        :param login: pararam login
        :param password: pararam password
        :param key: 16 chars second factor key to generate one time code
        :return: True if authentication success
        """
        key = key or self._key
        if not key:
            raise PararamioAuthenticationException('key must be set and not empty')
        return self._authenticate(do_second_step, key, login, password)

    def authenticate_with_code(self, code: str, login: str = None, password: str = None) -> bool:
        """
        Authenticate client with generated TFA code
        :param login: pararam login
        :param password: pararam password
        :param code: 6 digits code
        :return: True if authentication success
        """
        if not code:
            raise PararamioAuthenticationException('code must be set and not empty')
        return self._authenticate(do_second_step_with_code, code, login, password)

    def _api_request(self, url: str, method: str = 'GET', data: dict = None, callback: Callable = lambda rsp: rsp, raise_on_error: bool = False) -> Any:
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
            self._save_cookie()
        try:
            return callback(api_request(url, method, data, cookie_jar=self._cookie, headers=self.__headers))
        except PararamioHTTPRequestException as e:
            if raise_on_error:
                raise
            if e.code == 401:
                self._authenticated = False
                return self._api_request(url=url, method=method, data=data, callback=callback, raise_on_error=True)
            message = e.message
            if message == 'xsrf':
                log.info('xsrf is expire, invalid or was not set, trying to get new one')
                self.__headers[XSRF_HEADER_NAME] = ''
                return self._api_request(url=url, method=method, data=data, callback=callback, raise_on_error=True)
            raise

    def api_get(self, url: str, raise_on_error: bool = False) -> dict:
        return self._api_request(url, raise_on_error=raise_on_error)

    def api_post(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='POST', data=data, raise_on_error=raise_on_error)

    def api_put(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='PUT', data=data, raise_on_error=raise_on_error)

    def api_delete(self, url: str, data: Dict[Any, Any] = None, raise_on_error: bool = False) -> dict:
        return self._api_request(url, method='DELETE', data=data, raise_on_error=raise_on_error)

    def _upload_file(
        self,
        file: BinaryIO,
        chat_id: int,
        filename: str = None,
        type_: str = None,
        organization_id: int = None,
        reply_no: int = None,
        quote_range: str = None
    ) -> Tuple[dict, dict]:
        if type_ is None and not filename:
            raise PararamioValidationException('filename must be set when type is None')
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
        if type_ == 'organization_avatar' and organization_id is None:
            raise PararamioValidationException('organization_id must be set when type is organization_avatar')
        if type_ == 'chat_avatar' and chat_id is None:
            raise PararamioValidationException('chat_id must be set when type is chat_avatar')
        if type_ is None:
            if not mimetypes.inited:
                mimetypes.init(files=os.environ.get('PARARAMIO_MIME_TYPES_PATH', None))
            type_ = mimetypes.guess_type(filename)[0]
        file.seek(0, os.SEEK_END)
        file_size = file.tell()
        file.seek(0, 0)
        fields = [
            ('type', type_),
            ('filename', filename),
            ('size', file_size),
            ('chat_id', chat_id),
            ('organization_id', organization_id),
            ('reply_no', reply_no),
            ('quote_range', quote_range)
        ]
        return xupload_file(fp=file, fields=fields, headers=self.__headers, cookie_jar=self._cookie), dict(fields)

    def upload_file(
        self,
        file: Union[str, BytesIO, Path],
        chat_id: int,
        filename: str = None,
        reply_no: int = None,
        quote_range: str = None
    ) -> File:
        if isinstance(file, (str, Path)):
            filename = filename or os.path.basename(file)
            with open(file, 'rb') as f:
                res, extra = self._upload_file(file=f, chat_id=chat_id, filename=filename, reply_no=reply_no, quote_range=quote_range)
        else:
            res, extra = self._upload_file(file=file, chat_id=chat_id, filename=filename, reply_no=reply_no, quote_range=quote_range)
        return File(self, guid=res['guid'], mime_type=extra['type'], **extra)

    def delete_file(self, guid: str) -> dict:
        return delete_file(guid, headers=self.__headers, cookie_jar=self._cookie)

    def download_file(self, guid: str, filename: str) -> BytesIO:
        return download_file(guid, filename, headers=self.__headers, cookie_jar=self._cookie)

    @property
    def profile(self) -> dict:
        if not self.__profile:
            self.__profile = self._profile()
        return self.__profile

    def search_user(self, query: str) -> List[User]:
        return User.search(self, query)

    def search_group(self, query: str) -> List[Group]:
        return Group.search(self, query)

    def search_posts(self, query: str, order_type: str = 'time', page: int = 1, chat_id: int = None, limit: Optional[int] = None) -> Tuple[int, Iterable[Post]]:
        return Chat.post_search(self, query, order_type=order_type, page=page, chat_id=chat_id, limit=limit)

    def list_chats(self) -> Iterable:
        url = '/core/chat/sync'
        chats_per_load = 50
        ids = self.api_get(url).get('chats', [])
        return lazy_loader(self, ids, Chat.load_chats, per_load=chats_per_load)

    def post_private_message_by_user_email(self, email: str, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_email': email})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_id(self, user_id: int, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_id': user_id})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_unique_name(self, unique_name: str, text: str) -> Post:
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_unique_name': unique_name})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])



from typing import Any, Dict, List, TYPE_CHECKING


if TYPE_CHECKING:
    from pararamio.client import Pararamio

__all__ = ('Group',)


class Group:
    id: int
    _data: Dict[str, Any]
    _client: 'Pararamio'

    def __init__(self, client: 'Pararamio', id: int, **kwargs):
        self.id = id
        self._data = {}
        if kwargs:
            self._data = {'id': id, **kwargs}
        self._client = client

    def __getattr__(self, name: str) -> Any:
        return self._data[name]

    def __str__(self) -> str:
        title = self.title or 'New'
        id_ = self.id or ''
        return f'[{id_}] {title}'

    def __eq__(self, other) -> bool:
        if not isinstance(other, Group):
            return id(other) == id(self)
        return self.id == other.id

    def sync(self):
        # TODO: Implement sync
        raise NotImplementedError()

    @classmethod
    def search(cls, client: 'Pararamio', search_string: str) -> List['Group']:
        url = f'/users?flt={search_string}'
        return [cls(client, **group) for group in client.api_get(url).get('groups', [])]



from typing import Any, Callable, Dict, TYPE_CHECKING, Optional

from pararamio.exceptions import PararamioValidationException


if TYPE_CHECKING:
    from pararamio.client import Pararamio
ATTR_FORMATTERS: Dict[str, Callable[[dict, str], Any]] = {}


class File:
    """
    From /upload
    {
        "guid": "4faf5a45c5ba447ab7004b30181dd4c1"
    }
    From post.meta.file
    {
        "guid": "720ef462f19d43259f4af9fc2b75cdcf",
        "name": "image.png",
        "mime_type": "image/png",
        "size": 41935
    }
    """
    _client: 'Pararamio'
    _data: Dict[str, Any]
    guid: str

    def __init__(self, client, guid: str, **kwargs):
        self._client = client
        self.guid = guid
        self._data = {**kwargs, 'guid': guid}
        if 'name' in kwargs:
            self._data['filename'] = kwargs['name']

    def __getattr__(self, name):
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        return self._data[name]

    def __str__(self):
        return self._data.get('name', '')

    def serialize(self) -> Dict[str, str]:
        return self._data

    def delete(self):
        self._client.delete_file(self.guid)

    def download(self, filename: Optional[str] = None):
        if filename is None and 'filename' not in self._data:
            raise PararamioValidationException('can not determine filename')
        self._client.download_file(self.guid, filename or self._data['filename'])



import math
import uuid
import hashlib
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, cast
from html import unescape

from pararamio.constants import DATETIME_FORMAT
from pararamio.exceptions import PararamioValidationException


__all__ = (
    'encode_digit',
    'lazy_loader',
    'encode_chat_id',
    'join_ids',
    'parse_datetime',
    'parse_iso_datetime',
    'check_login_opts',
    'get_empty_vars',
    'unescape_dict',
)


def check_login_opts(login: Optional[str], password: Optional[str]) -> bool:
    return all(map(bool, [login, password]))


def get_empty_vars(**kwargs: Any):
    return ', '.join([k for k, v in kwargs.items() if not v])


def encode_digit(digit: int, res: str = '') -> str:
    if not isinstance(digit, int):
        digit = int(digit)
    code_string = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_.'
    result = math.floor(digit / len(code_string))
    res = code_string[int(digit % len(code_string))] + res
    return encode_digit(result, res) if result > 0 else res


def encode_chat_id(chat_id: int, posts_count: int, last_read_post_no: int) -> str:
    return '-'.join(map(str, [chat_id, posts_count, last_read_post_no]))


def encode_chats_ids(chats_ids: List[Tuple[int, int, int]]) -> str:
    return '/'.join(map(lambda *args: encode_chat_id(*args), chats_ids))


def lazy_loader(cls: Any, items: Sequence, load_fn: Callable[[Any, List], List], per_load: int = 50) -> Iterable:
    load_counter = 0
    loaded_items: List[Any] = []
    counter = 0

    def load_items():
        return load_fn(cls, items[(per_load * load_counter): (per_load * load_counter) + per_load])

    for _ in items:
        if not loaded_items:
            loaded_items = load_items()
        if counter >= per_load:
            counter = 0
            load_counter += 1
            loaded_items = load_items()
        yield loaded_items[counter]
        counter += 1


def join_ids(items: List[Any]) -> str:
    return ','.join(map(str, items))


def get_utc(date: datetime) -> datetime:
    if date.tzinfo is None:
        raise PararamioValidationException('is not offset-aware datetime')
    return cast(datetime, date - cast(timedelta, date.utcoffset()))


def parse_datetime(data: Dict[str, Any], key: str, format_: str = DATETIME_FORMAT) -> Optional[datetime]:
    if key not in data:
        return None
    return datetime.strptime(data[key], format_).replace(tzinfo=timezone.utc)


def parse_iso_datetime(data: Dict[str, Any], key: str) -> Optional[datetime]:
    try:
        return parse_datetime(data, key, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError:
        return parse_datetime(data, key, '%Y-%m-%dT%H:%M:%SZ')


def format_datetime(date: datetime) -> str:
    return get_utc(date).strftime(DATETIME_FORMAT)


def rand_id():
    _hash = hashlib.md5(bytes(uuid.uuid4().hex, 'utf8'))
    return str(int(int(_hash.hexdigest(), 16) * 10 ** -21))


def unescape_dict(d: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """
    Unescapes the html values of specified keys in a dictionary.

    :param d: The dictionary whose values are to be unescaped.
    :type d: Dict[str, Any]
    :param keys: The list of keys whose corresponding values in the dictionary are to be unescaped.
    :type keys: List[str]
    :return: A dictionary with the values of specified keys unescaped (other values remain unchanged).
    :rtype: Dict[str, Any]
    """
    return {k: unescape(v) if k in keys else v for k, v in d.items()}



import codecs
import json
import logging
import mimetypes
import os
from http.client import HTTPResponse
from io import BytesIO
from typing import BinaryIO, List, Optional, Tuple, Union
from urllib.error import HTTPError
from urllib.request import HTTPCookieProcessor, Request, build_opener

from pararamio._types import CookieJarT, HeaderLikeT
from pararamio.constants import BASE_API_URL, FILE_UPLOAD_URL, REQUEST_TIMEOUT as TIMEOUT, UPLOAD_TIMEOUT, VERSION
from pararamio.exceptions import PararamioHTTPRequestException


__all__ = ('api_request', 'bot_request', 'upload_file', 'xupload_file', 'delete_file', 'download_file', 'raw_api_request')
log = logging.getLogger('pararamio')
UA_HEADER = f'pararamio lib version {VERSION}'
DEFAULT_HEADERS = {
    'Content-type': 'application/json',
    'Accept':       'application/json',
    'User-agent':   UA_HEADER
}
writer = codecs.lookup('utf-8')[3]


def multipart_encode(fd: BinaryIO,
                     fields: List[Tuple[str, Union[str, None, int]]] = None,
                     boundary: str = None,
                     form_field_name: str = 'data',
                     filename: str = None,
                     content_type: str = None) -> bytes:
    if fields is None:
        fields = []
    if boundary is None:
        boundary = 'FORM-BOUNDARY'
    body = BytesIO()

    def write(text: str):
        nonlocal body
        writer(body).write(text)

    if fields:
        for (key, value) in fields:
            if value is None:
                continue
            write(f'--{boundary}\r\n')
            write(f'Content-Disposition: form-data; name="{key}"')
            write(f'\r\n\r\n{value}\r\n')
    if not filename:
        filename = os.path.basename(fd.name)
    if not content_type:
        content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
    fd.seek(0)
    write(f'--{boundary}\r\n')
    write(f'Content-Disposition: form-data; name="{form_field_name}"; filename="{filename}"\r\n')
    write(f'Content-Type: {content_type}\r\n\r\n')
    body.write(fd.read())
    write(f'\r\n--{boundary}--\r\n\r\n')
    return body.getvalue()


def bot_request(url: str, key: str, method: str = 'GET', data: Optional[dict] = None, headers: dict = None, timeout: int = TIMEOUT):
    _headers = {'X-APIToken': key, **DEFAULT_HEADERS}
    if headers:
        _headers = {**_headers, **headers}
    return api_request(url=url, method=method, data=data, headers=_headers, timeout=timeout)


def _base_request(url: str, method: str = 'GET', data: bytes = None, headers: dict = None, cookie_jar: CookieJarT = None,
                  timeout: int = TIMEOUT) -> HTTPResponse:
    _url = f'{BASE_API_URL}{url}'
    _headers = DEFAULT_HEADERS
    if headers:
        _headers = {**_headers, **headers}
    opener = build_opener(*[HTTPCookieProcessor(cookie_jar)] if cookie_jar is not None else [])
    _data = None
    if data:
        _data = data
    rq = Request(_url, _data, method=method, headers=_headers)
    log.debug('%s - %s - %s - %s - %s', _url, method, data, headers, cookie_jar)
    try:
        return opener.open(rq, timeout=timeout)
    except HTTPError as e:
        log.error('%s - %s - %s', _url, method, e)
        # noinspection PyUnresolvedReferences
        raise PararamioHTTPRequestException(e.filename, e.code, e.msg, e.hdrs, e.fp)  # type: ignore[attr-defined]


def _base_file_request(url: str, method='GET', data: bytes = None, headers: Optional[HeaderLikeT] = None, cookie_jar: CookieJarT = None,  # type: ignore
                       timeout: int = TIMEOUT) -> BytesIO:
    _url = f'{FILE_UPLOAD_URL}{url}'
    opener = build_opener(HTTPCookieProcessor(cookie_jar))
    if not headers:
        headers = {}
    rq = Request(_url, data, method=method, headers=headers)
    log.debug('%s - %s - %s - %s - %s', url, method, data, headers, cookie_jar)
    try:
        resp = opener.open(rq, timeout=timeout)
        if 200 >= resp.getcode() < 300:
            return resp
    except HTTPError as e:
        log.error('%s - %s - %s', _url, method, e)
        # noinspection PyUnresolvedReferences
        raise PararamioHTTPRequestException(e.filename, e.code, e.msg, e.hdrs, e.fp)  # type: ignore[attr-defined]


def upload_file(fp: BinaryIO,
                perm: str,
                filename: str = None,
                file_type=None,
                headers: HeaderLikeT = None,
                cookie_jar: CookieJarT = None,
                timeout: int = UPLOAD_TIMEOUT):
    url = f'/upload/{perm}'
    boundary = 'FORM-BOUNDARY'
    _headers = {
        'User-agent':   UA_HEADER,
        **(headers or {}),
        'Accept':       'application/json',
        'Content-Type': f'multipart/form-data; boundary={boundary}'
    }
    data = multipart_encode(fp, boundary=boundary, form_field_name='file', filename=filename, content_type=file_type)
    resp = _base_file_request(url, method='POST', data=data, headers=_headers, cookie_jar=cookie_jar, timeout=timeout)
    return json.loads(resp.read())


def xupload_file(fp: BinaryIO, fields: List[Tuple[str, Union[str, None, int]]], headers: HeaderLikeT = None, cookie_jar: CookieJarT = None,
                 timeout: int = UPLOAD_TIMEOUT) -> dict:
    url = '/upload'
    boundary = 'FORM-BOUNDARY'
    _headers = {
        'User-agent':   UA_HEADER,
        **(headers or {}),
        'Accept':       'application/json',
        'Content-Type': f'multipart/form-data; boundary={boundary}'
    }
    data = multipart_encode(fp, fields, boundary=boundary)
    resp = _base_file_request(url, method='POST', data=data, headers=_headers, cookie_jar=cookie_jar, timeout=timeout)
    return json.loads(resp.read())


def delete_file(guid: str, headers: HeaderLikeT = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> dict:
    url = f'/delete/{guid}'
    resp = _base_file_request(url, method='DELETE', headers=headers, cookie_jar=cookie_jar, timeout=timeout)
    return json.loads(resp.read())


def download_file(guid: str, filename: str, headers: HeaderLikeT = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> BytesIO:
    url = f'/download/{guid}/{filename}'
    res = file_request(url, method='GET', headers=headers, cookie_jar=cookie_jar, timeout=timeout)
    return res


def file_request(url: str, method='GET', data: bytes = None, headers: HeaderLikeT = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> BytesIO:
    _headers = DEFAULT_HEADERS
    if headers:
        _headers = {**_headers, **headers}
    return _base_file_request(url, method=method, data=data, headers=_headers, cookie_jar=cookie_jar, timeout=timeout)


def raw_api_request(url: str,
                    method: str = 'GET',
                    data: dict = None,
                    headers: HeaderLikeT = None,
                    cookie_jar: CookieJarT = None,
                    timeout: int = TIMEOUT) -> Tuple[dict, List]:
    resp = _base_request(**locals())
    if 200 >= resp.getcode() < 300:
        contents = resp.read()
        return json.loads(contents), resp.getheaders()
    return {}, []


def api_request(url: str, method: str = 'GET', data: Optional[dict] = None, headers: HeaderLikeT = None, cookie_jar: CookieJarT = None,
                timeout: int = TIMEOUT) -> dict:
    _data = None
    if data is not None:
        _data = str.encode(json.dumps(data), 'utf-8')
    resp = _base_request(url, method, _data, headers, cookie_jar, timeout)
    if 200 <= resp.getcode() < 500:
        return json.loads(resp.read())
    return {}



from collections import OrderedDict
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, TYPE_CHECKING
from pararamio.exceptions import PararamioRequestException, PararamNoNextPost, PararamNoPrevPost
from pararamio.file import File
from pararamio.utils.helpers import encode_digit, parse_iso_datetime, rand_id

if TYPE_CHECKING:
    from pararamio.chat import Chat
    from pararamio.client import Pararamio
    from pararamio._types import PostMention, PostMetaT, TextParsedT

__all__ = ('Post',)
ATTR_FORMATTERS = {
    'post_no': lambda data, key: int(data[key]),
    'time_edited': parse_iso_datetime,
    'time_created': parse_iso_datetime,
}


class Post:
    __slots__ = ('_data', '_chat', 'post_no', 'load_on_key_error')
    _data: Dict[str, Any]
    _chat: 'Chat'
    post_no: int
    is_bot: bool
    text: str
    text_parsed: Optional[List['TextParsedT']]
    user_id: int
    event: Optional[Dict[str, Any]]
    time_created: datetime
    time_edited: Optional[datetime]
    reply_no: Optional[int]
    meta: 'PostMetaT'
    uuid: Optional[str]
    load_on_key_error: bool

    def __init__(self, chat: 'Chat', post_no: int = None, load_on_key_error: bool = True, **kwargs):
        self._chat = chat
        if post_no is None:
            post_no = kwargs['in_thread_no']
        self.post_no = post_no
        self._data = {**kwargs, 'post_no': post_no}
        self.load_on_key_error = load_on_key_error

    def __getattr__(self, name):
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        try:
            return self._data[name]
        except KeyError:
            if self.load_on_key_error:
                self.load()
                return self._data[name]
            raise

    def __repr__(self):
        return f'<Post(client={hex(id(self.client))}, chat_id={self.chat_id}, post_no={self.post_no}) {hex(id(self))}>'

    def __str__(self):
        return self.text

    def __eq__(self, other):
        if not isinstance(other, Post):
            return id(other) == id(self)
        return self._chat == other._chat and self.post_no == other.post_no

    def _compare_validations(self, other):
        if not isinstance(other, Post):
            raise ValueError('can not compare post and {}'.format(other.__class__.__name__))
        if self._chat != other._chat:
            raise ValueError('can not compare posts from different chats')

    def __ge__(self, other: 'Post'):
        self._compare_validations(other)
        # noinspection PyUnresolvedReferences
        return self.post_no >= other.post_no  # type: ignore[operator]

    def __gt__(self, other: 'Post'):
        self._compare_validations(other)
        # noinspection PyUnresolvedReferences
        return self.post_no > other.post_no  # type: ignore[operator]

    def __lt__(self, other: 'Post'):
        self._compare_validations(other)
        # noinspection PyUnresolvedReferences
        return self.post_no < other.post_no  # type: ignore[operator]

    def __le__(self, other: 'Post'):
        self._compare_validations(other)
        # noinspection PyUnresolvedReferences
        return self.post_no <= other.post_no  # type: ignore[operator]

    @property
    def chat_id(self) -> int:
        return self._chat.id

    @property
    def in_thread_no(self) -> int:
        return self.post_no

    @property
    def file(self):
        _file = self.meta.get('file', None)
        if not _file:
            return
        return File(self._chat._client, **_file)

    @property
    def is_reply(self) -> bool:
        return self.reply_no is not None

    @property
    def is_deleted(self) -> bool:
        return bool(self.text)

    @property
    def chat(self) -> 'Chat':
        return self._chat

    @property
    def client(self) -> 'Pararamio':
        return self._chat._client

    @property
    def is_bot(self) -> bool:
        return self.meta.get('user', {}).get('is_bot', False)

    @property
    def is_file(self) -> bool:
        return 'file' in self.meta

    @property
    def is_event(self) -> bool:
        """
        is this message is event
        """
        return bool(self.event)

    @property
    def mentions(self) -> List['PostMention']:
        text_parsed = getattr(self, 'text_parsed', None)
        if not text_parsed:
            return []
        mentions_ = []
        for item in text_parsed:
            if item.get('type', '') == 'mention':
                mentions_.append(
                    {
                        'id': item.get('id', None),
                        'name': item.get('name', None),
                        'value': item.get('value', None)
                    }
                )
        return mentions_

    @property
    def user_links(self) -> List['PostMention']:
        text_parsed = getattr(self, 'text_parsed', None)
        if not text_parsed:
            return []
        links = []
        for item in text_parsed:
            if item.get('type', '') == 'user_link':
                links.append(
                    {
                        'id': item.get('id', None),
                        'name': item.get('name', None),
                        'value': item.get('value', None)
                    }
                )
        return links

    @property
    def is_mention(self) -> bool:
        text_parsed = getattr(self, 'text_parsed', None)
        if not text_parsed:
            return False
        for item in text_parsed:
            if item.get('type', '') == 'mention' and item.get('id', None) == self.client.profile.get('id', -1):
                return True
        return False

    def load(self) -> 'Post':
        url = f'/msg/post?ids={encode_digit(self.chat.id)}-{encode_digit(self.post_no)}'  # type: ignore
        res = self.client.api_get(url).get('posts', [])
        if len(res) != 1:
            raise PararamioRequestException(f'failed to load data for post_no {self.post_no} in chat {self._chat.id}')
        self._data = res[0]
        return self

    @property
    def replies(self):
        url = f'/msg/post/{self._chat.id}/{self.post_no}/replies'
        return self.client.api_get(url)

    def reply(self, text: str, quote: Optional[str] = None) -> 'Post':
        _url = f'/amsg/post/{self._chat.id}'
        res = self.client.api_post(_url, {
            'uuid':     rand_id(),
            'text':     text,
            'quote':    quote,
            'reply_no': self.post_no
        })
        return Post(self._chat, res['post_no'], load_on_key_error=self.load_on_key_error).load()

    def rerere(self) -> Iterable['Post']:
        url = f'/msg/post/{self._chat.id}/{self.post_no}/rerere'
        res = self.client.api_get(url)

        def make_post_from_re(post_no):
            return Post(self._chat, post_no, load_on_key_error=self.load_on_key_error).load()

        return map(make_post_from_re, res['data'])

    def get_tree(self, load_limit: int = 1000) -> 'OrderedDict[int, Post]':
        posts = {self.post_no: self}
        for post in self.rerere():
            posts[post.post_no] = post
        first = posts[min(posts.keys())]
        tree = OrderedDict(sorted(posts.items()))  # type: ignore
        load_start = first.post_no + 1
        if self.post_no - first.post_no > load_limit:
            load_start = self.post_no - load_limit
        for post in self.chat._lazy_posts_loader(*sorted([load_start, self.post_no - 1])):
            posts[post.post_no] = post

        for no, post in sorted(posts.items()):
            if post.reply_no is None or post.reply_no not in tree:
                continue
            tree[post.post_no] = post
        return OrderedDict(sorted(tree.items()))  # type: ignore

    def get_reply_to_post(self) -> Optional['Post']:
        reply_no = self.reply_no
        if reply_no is not None:
            return Post(self._chat, reply_no, load_on_key_error=self.load_on_key_error).load()
        return None

    def next(self, skip_event: bool = True) -> 'Post':
        """
        get next post or throw PararamNoNextPost

        :param bool skip_event: Skip message if this is an event
        """
        _next = self.post_no + 1
        if _next > self._chat.posts_count:
            raise PararamNoNextPost()
        post = Post(self._chat, _next)
        if skip_event and post.is_event:
            return post.next()
        return post

    def prev(self, skip_event: bool = True) -> 'Post':
        """
        get previous post or throw PararamNoPrevPost

        :param bool skip_event: Skip message if this is an event
        """
        _prev = self.post_no - 1
        if _prev <= 0:
            raise PararamNoPrevPost()
        post = Post(self._chat, _prev)
        if skip_event and post.is_event:
            return post.prev()
        return post

    def who_read(self) -> List[int]:
        url = f'/activity/who-read?thread_id={self._chat.id}&post_no={self.post_no}'
        return self.client.api_get(url).get('users', [])

    def edit(self, text: str, quote: Optional[str] = None, reply_no: int = None) -> bool:
        """
        :param quote: str, used with reply_no, default None
        :param reply_no: int reply to, default None
        :param text: str Text to change
        :return: Post
        """
        url = f'/amsg/post/{self._chat.id}/{self.post_no}'

        res = self.client.api_put(url, {
            'uuid':     self._data.get('uuid', rand_id()),
            'text':     text,
            'quote':    quote,
            'reply_no': reply_no
        })
        if res.get('ver'):
            self.load()
            return True
        return False

    def delete(self) -> bool:
        url = f'/amsg/post/{self._chat.id}/{self.post_no}'
        res = self.client.api_delete(url)
        if res.get('ver'):
            self.load()
            return True
        return False

    @classmethod
    def create(cls, chat: 'Chat', text: str, reply_no: Optional[int] = None, quote: Optional[str] = None) -> 'Post':
        url = f'/amsg/post/{chat.id}'
        res = chat._client.api_post(url, {
            'uuid':  rand_id(), 'text': text,
            'quote': quote, 'reply_no': reply_no
        })

        if not res:
            raise PararamioRequestException('Failed to create post')

        return cls(chat, post_no=res['post_no']).load()



VERSION = '2.0.2'
BASE_API_URL = 'https://api.pararam.io'
FILE_UPLOAD_URL = 'https://file.pararam.io'
DEFAULT_CONNECTION_NAME = '_default'
DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
XSRF_HEADER_NAME: str = 'X-Xsrftoken'
UPLOAD_TIMEOUT = 600
REQUEST_TIMEOUT = 30
POSTS_LIMIT = 100



from .chat import Chat
from .client import Pararamio
from .deferred_post import DeferredPost
from .group import Group
from .post import Post
from .team import Team, TeamMember
from .user import User
from .bot import PararamioBot
from .activity import Activity, ActivityAction
from .poll import Poll


__all__ = (
    'Pararamio',
    'Chat',
    'Post',
    'DeferredPost',
    'User',
    'Group',
    'Team',
    'TeamMember',
    'PararamioBot',
    'ActivityAction',
    'Activity',
    'Poll',
)



import base64
import tkinter as tk
from urllib.error import HTTPError
from urllib.parse import urlencode

from pararamio._types import CookieJarT
from pararamio.constants import REQUEST_TIMEOUT as TIMEOUT
from pararamio.utils.requests import _base_request, api_request


def get_captcha_img(id_: str, headers: dict = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> 'tk.PhotoImage':
    args = urlencode({'id': id_})
    url = f'/auth/captcha?{args}'
    tc = 3
    data = None
    while True:
        try:
            data = _base_request(url, headers=headers, cookie_jar=cookie_jar, timeout=timeout).read()
            break
        except HTTPError:
            if not tc:
                raise
            tc -= 1
            continue
    return tk.PhotoImage(data=base64.b64encode(data))


def verify_captcha(code: str, headers: dict = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> bool:
    url = '/auth/captcha'
    resp = api_request(url, 'POST', data={'code': code}, headers=headers, cookie_jar=cookie_jar, timeout=timeout)
    return str.lower(resp.get('status', '')) == 'OK'


def show_captcha(login: str, headers: dict = None, cookie_jar: CookieJarT = None, timeout: int = TIMEOUT) -> bool:
    root = tk.Tk()
    root.title('Pararamio captcha')
    root.resizable(False, False)
    photo = get_captcha_img(login, headers=headers, cookie_jar=cookie_jar, timeout=timeout)
    h = photo.height() + 138
    w = photo.width() + 75
    root.geometry(f'{w:d}x{h:d}')
    panel = tk.Label(root, image=photo)
    br = tk.Button(text='R', width=7, height=4, cursor='hand2')
    bs = tk.Button(text='Send', width=15, height=4, cursor='hand2')
    label = tk.Label(root, text='Code: ', width=20)
    code = tk.StringVar()
    entry = tk.Entry(root, textvariable=code, justify='center')
    cookie_entered = False

    def refresh_photo():
        # noinspection PyBroadException
        try:
            img = get_captcha_img(login, headers=headers, cookie_jar=cookie_jar, timeout=timeout)
            panel.config(image=img)
            panel.image = img
        except Exception:
            pass

    def send_code():
        nonlocal cookie_entered
        try:
            verify_captcha(str.strip(entry.get()), headers=headers, cookie_jar=cookie_jar, timeout=timeout)
            cookie_entered = True
            root.destroy()
        except HTTPError:
            code.set('')
            refresh_photo()

    br.config(command=refresh_photo)
    bs.config(command=send_code)
    panel.grid(row=0, column=0, pady=0, padx=0, sticky=tk.N + tk.S + tk.W + tk.E)
    br.grid(row=0, column=1, pady=0, padx=0, sticky=tk.N + tk.S + tk.W + tk.E)
    label.grid(row=1, pady=0, padx=0, columnspan=2, sticky=tk.N + tk.S + tk.W + tk.E)
    entry.grid(row=2, pady=0, padx=0, columnspan=2, sticky=tk.N + tk.S + tk.W + tk.E)
    bs.grid(row=3, columnspan=2)
    root.mainloop()
    return cookie_entered



from typing import Any, Dict, List, Optional, TYPE_CHECKING

from pararamio.client import Pararamio
from pararamio.exceptions import PararamioRequestException
from pararamio.user import User
from pararamio.utils import parse_iso_datetime


if TYPE_CHECKING:
    from datetime import datetime
__all__ = ('TeamMember', 'Team', 'TeamMemberStatus', )

ATTR_FORMATTERS = {
    'time_edited':  parse_iso_datetime,
    'time_created': parse_iso_datetime,
}
MEMBER_ATTR_FORMATTERS = {
    **ATTR_FORMATTERS,
    'last_activity': parse_iso_datetime
}
STATUS_ATTR_FORMATTERS = {
    'time_created': parse_iso_datetime,
}


class TeamMemberStatus:
    __slots__ = ('client', '_data', )
    client: 'Pararamio'
    _data: Dict[str, Any]
    id: int
    user_id: int
    setter_id: int
    org_id: int
    time_created: 'datetime'
    status: str

    def __init__(self, client: 'Pararamio', id: int, **kwargs):
        self.client = client
        self._data = {**kwargs, 'id': int(id)}

    def __getattr__(self, name):
        fmt_fnc = STATUS_ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        return self._data[name]


class TeamMember:
    __slots__ = ('client', '_data')
    _data: Dict[str, Any]
    chats: List[int]
    email: str
    groups: List[int]
    id: int
    org_id: int
    inviter_id: Optional[int]
    is_admin: bool
    is_member: bool
    last_activity: 'datetime'
    phonenumber: str
    state: str
    time_created: 'datetime'
    time_updated: 'datetime'
    two_step_enabled: bool

    def __init__(self, client: 'Pararamio', id: int, org_id: int, **kwargs):
        self.client = client
        self._data = {**kwargs, 'id': int(id), 'org_id': int(org_id)}

    def __getattr__(self, name):
        fmt_fnc = MEMBER_ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        return self._data[name]

    def __str__(self):
        text = self._data.get('text', None)
        if text is None:
            self.load()
            text = self._data['text']
        return text

    def __eq__(self, other):
        if not isinstance(other, (TeamMember, User)):
            return hash(other) == hash(self)
        return self.id == other.id

    @property
    def user(self) -> User:
        return User(client=self.client, id=self.id)

    def get_last_status(self) -> Optional['TeamMemberStatus']:
        url = f'/core/org/status?user_ids={self.id}'
        res = self.client.api_get(url).get('data', [])
        if not res:
            return None
        return TeamMemberStatus(self.client, **res[0])

    def add_status(self, status: str) -> bool:
        url = f'/core/org/status'
        data = {
            'org_id': self.org_id,
            'status': status,
            'user_id': self.id,
        }
        res = self.client.api_post(url, data=data)
        return bool(res) and res.get('result') == 'OK'


class Team:
    __slots__ = ('_data', 'client', 'load_on_key_error',)
    _data: Dict[str, Any]
    client: 'Pararamio'
    admins: List[int]
    default_thread_id: int
    description: Optional[str]
    email_domain: Optional[str]
    groups: List[int]
    guest_thread_id: Optional[int]
    guests: List[int]
    id: int
    inviter_id: Optional[int]
    is_active: bool
    slug: str
    state: str
    time_created: str
    time_updated: str
    title: str
    two_step_required: bool
    users: List[int]
    load_on_key_error: bool

    def __init__(self, client: 'Pararamio', id: int, load_on_key_error: bool = True, **kwargs):
        self.client = client
        self.load_on_key_error = load_on_key_error
        self._data = {**kwargs, 'id': int(id)}

    def __getattr__(self, name):
        fmt_fnc = ATTR_FORMATTERS.get(name, None)
        if fmt_fnc:
            return fmt_fnc(self._data, name)
        try:
            return self._data[name]
        except KeyError:
            if self.load_on_key_error:
                self.load()
                return self._data[name]
            raise

    def __str__(self):
        text = self._data.get('text', None)
        if text is None:
            self.load()
            text = self._data['text']
        return text

    def __eq__(self, other):
        if not isinstance(other, Team):
            return hash(other) == hash(self)
        return self.id == other.id

    def __contains__(self, item):
        if not isinstance(item, (TeamMember, User)):
            return False
        return item.id in self.users

    def load(self):
        url = f'/core/org?ids={self.id}'
        res = self.client.api_get(url)
        self._data.update(res)
        return self

    def member_info(self, user_id: int) -> 'TeamMember':
        url = f'/core/org/{self.id}/member_info/{user_id}'
        res = self.client.api_get(url)
        if not res:
            raise PararamioRequestException('empty response for user %s', user_id)
        return TeamMember(self.client, org_id=self.id, **res)

    def members_info(self) -> List['TeamMember']:
        url = f'/core/org/{self.id}/member_info'
        res = self.client.api_get(url)
        if res:
            return [TeamMember(self.client, org_id=self.id, **m) for m in res.get('data', [])]
        return []

    @classmethod
    def my_team_ids(cls, client: 'Pararamio') -> List[int]:
        url = '/core/org/sync'
        res = client.api_get(url) or {}
        return res.pop('ids', [])

    @classmethod
    def load_teams(cls, client: 'Pararamio') -> List['Team']:
        ids = cls.my_team_ids(client)

        if ids:
            ids = ','.join(str(x) for x in ids)
            url = f'/core/org?ids={ids}'
            res = client.api_get(url)

            if res:
                return [cls(client, **r) for r in res['orgs']]

        return []
